// consolemain.c
//
// Time-stamp: <09/02/02 17:30:11 keuchel@netwave.de>

#include <stdlib.h>
#if UNDER_CE > 201
#include <stdio.h>
#endif
#include "w32console.h"
#include "celib.h"
#include "ndir.h"
#include "sys/stat.h"
#include "glob.h"
#include "getopt.h"
#include <sys/timeb.h>
#include <time.h>

#ifdef USE_READLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif

static int a_argc;
static char **a_argv;
static int a_size;

extern HWND hMainWnd;
 
int argv_setargv(char *command);
void argv_freeargv();

char lastcommand[1024];
int interactive = 1;
char *progname;
int linecount = 1;
int hidewindow;
char *batchpath;
int globflag = 1;

char *readline(char *prompt);

#define MAXVARS  20
#define MAXPROGS 30

#define ISVARNAME(C) (isalpha(C) || isdigit(C) || (C) == '_')

struct varentry {
  char *name;
  char *value;
};

#define PROG_BUILTIN  1
#define PROG_PIPEABLE 2

struct progentry {
  char *name;
  int flags;
};

struct varentry vartab[MAXVARS];
struct progentry progtab[MAXPROGS];

struct varentry *
findvar(const char *name)
{
  int i;

  for(i = 0; i < MAXVARS; i++)
    {
      if(vartab[i].name && (strcmp(vartab[i].name, name) == 0))
	{
	  return &vartab[i];
	}
    }
  return NULL;
}

struct progentry *
findprog(const char *name)
{
  int i;

  for(i = 0; i < MAXPROGS; i++)
    {
      if(progtab[i].name && (strcmp(progtab[i].name, name) == 0))
	{
	  return &progtab[i];
	}
    }
  return NULL;
}

int 
addvar(const char *name, const char *val)
{
  int i;
  struct varentry *v;

  if((v = findvar(name)) != NULL)
    {
      free(v->value);
      v->value = strdup(val);
      return 0;
    }

  for(i = 0; i < MAXVARS; i++)
    {
      if(vartab[i].name == NULL)
	{
	  vartab[i].name = strdup(name);
	  vartab[i].value = strdup(val);
	  return 0;
	}
    }

  XCEShowMessageA("Vartable overflow (%d)", MAXVARS);
  return -1;
}

int 
addprog(const char *name, int flags)
{
  int i;
  struct progentry *v;

  for(i = 0; i < MAXPROGS; i++)
    {
      if(progtab[i].name == NULL)
	continue;

      if(strcmp(progtab[i].name, name) == 0)
	{
	  progtab[i].flags |= flags;
	  return 0;
	}
    }

  for(i = 0; i < MAXPROGS; i++)
    {
      if(progtab[i].name == NULL)
	{
	  progtab[i].name = strdup(name);
	  progtab[i].flags = flags;
	  return 0;
	}
    }

  XCEShowMessageA("Progtable overflow (%d)", MAXPROGS);
  return -1;
}

// no buffer checks...

void
substitutevars(char *buf)
{
  char newbuf[512];
  char *s, *d;
  struct varentry *v;
  char varname[126];
  char *p;

  for(s = buf, d = newbuf; *s;)
    {
      if(*s == '$')
	{
	  p = varname;
	  s++;
	  while(ISVARNAME(*s))
	    *p++ = *s++;
	  *p = 0;
	  if((v = findvar(varname)) != NULL)
	    {
	      strcpy(d, v->value);
	      d += strlen(v->value);
	    }
	}
      else
	{
	  *d++ = *s++;
	}
    }

  *d = 0;

  strcpy(buf, newbuf);
}

int
is_pipeable(const char *cmd)
{
  struct progentry *p;

  if((p = findprog(cmd)) == NULL)
    return 0;
  
  if(p->flags & PROG_PIPEABLE)
    return 1;

  return 0;
}

int
is_builtin(const char *cmd)
{
  struct progentry *p;

  if((p = findprog(cmd)) == NULL)
    return 0;
  
  if(p->flags & PROG_BUILTIN)
    return 1;

  return 0;
}

void
test1()
{
  char buf[126];

  struct _timeb tb;
  ftime(&tb);

  //printf("%.24s Millis: %d\n", ctime(&tb.time), tb.millitm);

  printf("\033[3;1H\033[7mName:   \033[0m\033[K");
  printf("\033[4;1H\033[7mStreet: \033[0m\033[K");
  printf("\033[5;1H\033[7mCity:   \033[0m\033[K");

  printf("\033[3;10H");
  fgets(buf, sizeof(buf), stdin);
  printf("\033[4;10H");
  fgets(buf, sizeof(buf), stdin);
  printf("\033[5;10H");
  fgets(buf, sizeof(buf), stdin);

  printf("\033[H\033[1;6r");
  printf("\033[L");
  printf("\033[L");
};

void
tabtest()
{
  printf("TabTest\n");
  printf("A\tB\n");
  printf("AB\tBC\n");
  printf("ABC\tBCD\n");
  printf("ABCD\tBCDE\n");
  printf("ABCDE\tBCDEF\n");
  printf("ABCDEF\tBCDEFG\n");
  printf("ABCDEFG\tBCDEFGH\n");
  printf("ABCDEFGH\tBCDEFGHI\n");
  printf("ABCDEFGHI\tBCDEFGHIJ\n");
  printf("ABCDEFGHIJ\tBCDEFGHIJ\n");
}

void
showversion()
{
  printf("\033[2J");
  printf("\033[7m");
  printf("ConsoleCE Version 1.3 by Rainer Keuchel");
  printf("\033[0m");
  printf("\n");
#ifdef USE_READLINE
  printf("Using GNU readline-4.2\n");
#endif
}

int
handle_command(int argc, char **argv)
{
  char buf[256];

  if(strcmp(argv[0], "pipeable") == 0)
    {
      int i;

      for(i = 1; i < argc; i++)
	{
	  addprog(argv[i], PROG_PIPEABLE);
	}
    }
  else if(strcmp(argv[0], "builtin") == 0)
    {
      int i;

      for(i = 1; i < argc; i++)
	{
	  addprog(argv[i], PROG_BUILTIN);
	}
    }
  else if(strcmp(argv[0], "echo") == 0)
    {
      char *str, *newstr;
      int nonewline = 0;
      char *s, *d;
      if(argc == 2)
	{
	  str = argv[1];
	}
      else if(argc == 3 && strcmp(argv[1], "-n") == 0)
	{
	  nonewline = 1;
	  str = argv[2];
	}
      else
	{
	  fprintf(stderr, "Usage: echo [-n ] <string>\n");
	  return -1;
	}

      newstr = alloca(strlen(str) + 1);

      for(s = str, d = newstr; *s;)
	{
	  if(*s == '\\')
	    {
	      if(s[1] == 'n')
		{
		  *d++ = '\n';
		  s += 2;
		}
	      else if(isdigit(s[1]) && isdigit(s[2]) && isdigit(s[3]))
		{
		  int c;
		  sscanf(s + 1, "%o", &c);
		  *d++ = c;
		  s += 4;
		}
	      else
		{
		  *d++ = *s++;
		}
	    }
	  else
	    {
	      *d++ = *s++;
	    }
	}
      *d = 0;

      printf("%s", newstr); 

      if(!nonewline)
	printf("\n");

      fflush(stdout);
    }
  else if(strcmp(argv[0], "set") == 0)
    {
      if(argc == 1)
	{
	  int i;

	  for(i = 0; i < MAXVARS; i++)
	    {
	      if(vartab[i].name)
		printf("%s=%s\n", vartab[i].name, vartab[i].value);
	    }
	  return 0;
	}
      else if(argc == 4)
	{
	  if(strcmp(argv[2], "=") != 0)
	    goto usage;
	  return addvar(argv[1], argv[3]);
	}
    usage:
      fprintf(stderr, "Usage: set [<varname> = <value>]\n");
      return -1;
    }
  else if(strcmp(argv[0], "read") == 0)
    {
      char buf[512];
      char *p;

      if(argc != 2)
	{
	  fprintf(stderr, "Usage: read <varname>\n");
	  return -1;
	}

      if(cgets(buf) == NULL)
	{
	  XCEShowMessageA("Cannot read from console");
	  return -1;
	}

      if((p = strrchr(buf, '\n')) != NULL)
	*p = 0;

      return addvar(argv[1], buf);
    }
  else if(strcmp(argv[0], "env") == 0)
    {
      char *p;

      if(argc != 2)
	{
	  fprintf(stderr, "Usage: env <varname>\n");
	  return -1;
	}

      if((p = getenv(argv[1])) == NULL)
	{
	  fprintf(stderr, "%s is not set\n", argv[0]);
	  return -1;
	}
      printf("%s=%s\n", argv[1], p);
    }
  else if(strcmp(argv[0], "setenv") == 0)
    {
      char buf[1024];

      if(argc != 3)
	{
	  fprintf(stderr, "Usage: setenv <varname> <value>\n");
	  return -1;
	}

      XCESetEnvironmentVariableInRegA(argv[1], argv[2]);
      sprintf(buf, "%s=%s", argv[1], argv[2]);
      putenv(buf);
    }
  else if(strcmp(argv[0], "pwd") == 0)
    {
      getcwd(buf, sizeof(buf));
      printf("%s\n", buf);
    }
  else if(strcmp(argv[0], "cd") == 0)
    {
      char *path;
      char buf[256];

      if(argc == 1)
	{
	  path = getenv("HOME");
	  if(path == NULL)
	    path = "/";
	}
      else if(argc == 2)
	{
	  path = argv[1];
	}
      else
	{
	  fprintf(stderr, "Usage: cd [<dir>]\n");
	  return -1;
	}

      if(chdir(path) < 0)
	{
	  perror(path);
	  return -1;
	}

      getcwd(buf, sizeof(buf));

      printf("Current dir now %s\n", buf);
    }
  else if(strcmp(argv[0], "mkdir") == 0 && is_builtin(argv[0]))
    {
      if(argc != 2)
	{
	  fprintf(stderr, "Usage: mkdir <dir>\n");
	  return -1;
	}

      if(mkdir(argv[1]) < 0)
	{
	  perror(argv[1]);
	  return -1;
	}
    }
  else if(strcmp(argv[0], "rm") == 0 && is_builtin(argv[0]))
    {
      int i;

      if(argc < 2)
	{
	  fprintf(stderr, "Usage: rm <file> [...]\n");
	  return -1;
	}

      for(i = 1; i < argc; i++)
	{
	  if(remove(argv[i]) < 0)
	    {
	      perror(argv[i]);
	      return -1;
	    }
	}
    }
  else if(strcmp(argv[0], "rmdir") == 0 && is_builtin(argv[0]))
    {
      if(argc != 2)
	{
	  fprintf(stderr, "Usage: rmdir <dirname>\n");
	  return -1;
	}

      if(rmdir(argv[1]) < 0)
	{
	  perror(argv[1]);
	  return -1;
	}
    }
  else if(strcmp(argv[0], "mv") == 0 && is_builtin(argv[0]))
    {
      if(argc != 3)
	{
	  fprintf(stderr, "Usage: mv <oldfile> <newfile>\n");
	  return -1;
	}

      if(rename(argv[1], argv[2]) < 0)
	{
	  fprintf(stderr, "Cannot rename %s to %s: %s\n", argv[1], argv[2], 
		  strerror(errno));
	  return -1;
	}
    }
  else if(strcmp(argv[0], "cp") == 0 && is_builtin(argv[0]))
    {
      if(argc != 3)
	{
	  fprintf(stderr, "Usage: cp <oldfile> <newfile>\n");
	  return -1;
	}

      if(XCECopyFileA(argv[1], argv[2], TRUE) < 0)
	{
	  fprintf(stderr, "Cannot copy %s to %s: %s\n", argv[1], argv[2], 
		  strerror(errno));
	  return -1;
	}
    }
  else if(strcmp(argv[0], "x") == 0 ||
	  strcmp(argv[0], "start") == 0 ||
	  strcmp(argv[0], "exec") == 0)
    {
      int i;
      char *cmdline = NULL;
      int retcode;
      DWORD dwProcId;
      BOOL bWait = TRUE;

      if(argc < 2)
	{
	  fprintf(stderr, "Usage: start <commandline>\n");
	  return -1;
	}

      cmdline = malloc(10240);
      cmdline[0] = 0;

      // FIXME: Overflow...
      for(i = 1; i < argc; i++)
	{
	  strcat(cmdline, "\"");
	  strcat(cmdline, argv[i]);
	  strcat(cmdline, "\" ");
	}

      if(strcmp(argv[0], "start") == 0)
	bWait = FALSE;

      retcode = XCEExecuteProcessA(cmdline, bWait, &dwProcId);

      free(cmdline);

      if(retcode < 0)
	{
	  DWORD dwError;

	  dwError = GetLastError();

	  if(dwError == ERROR_FILE_NOT_FOUND)
	    fprintf(stderr, "%s: file not found\n", argv[0]);
	  else
	    fprintf(stderr, "OsError: %d\n", dwError);
	  return -1;
	}
      else if(retcode != 0)
	{
	  fprintf(stderr, "Retcode: %d\n", retcode);
	  return -1;
	}

      SetForegroundWindow(hMainWnd);
      SetActiveWindow(hMainWnd);
    }
  else if(strcmp(argv[0], "dir") == 0 && is_builtin(argv[0]))
    {
      DIR *dirp;
      struct direct *dirent;
      struct stat st;
      char *opts;
      char *dirname;
      int oind = 1;

      if(argc > 3)
	{
	  fprintf(stderr, "Usage: dir [<opts>] [<dir>]\n");
	  return -1;
	}

      if(argc > 1 && argv[oind][0] == '-')
	{
	  opts = argv[oind++];
	}
      else
	{
	  opts = "";
	}

      if(argc > 1 && argv[oind])
	{
	  dirname = argv[oind++];
	}
      else
	{
	  dirname = ".";
	}

      if((dirp = opendir(dirname)) == NULL)
	{
	  perror(dirname);
	  return -1;
	}

      while((dirent = readdir(dirp)))
	{
	  char path[MAX_PATH];

	  if(strcmp(dirent->d_name, ".") == 0 ||
	     strcmp(dirent->d_name, "..") == 0)
	    continue;

	  sprintf(path, "%s/%s", dirname, dirent->d_name);
	  if(stat(path, &st) < 0)
	    perror(path);

	  printf("%s", dirent->d_name);
	  if(strchr(opts, 'l'))
	    {
	      printf(" %d", st.st_size);
	    }
	  printf("\n");
	}

      closedir(dirp);
    }
  else if(strcmp(argv[0], "ls") == 0 && is_builtin(argv[0]))
    {
      struct stat st;
      char *opts;
      char *pattern;
      int oind = 1;
      int globflags;
      glob_t gbuf;
      int i;

      globflags = (GLOB_ERR | GLOB_NOCHECK | GLOB_NOESCAPE | GLOB_NOMAGIC);

      gbuf.gl_pathc = 0;
      gbuf.gl_offs = 0;
      gbuf.gl_pathv = NULL;

      if(argc > 3)
	{
	  fprintf(stderr, "Usage: ls [<opts>] [<pattern>]\n");
	  return -1;
	}

      if(argc > 1 && argv[oind][0] == '-')
	{
	  opts = argv[oind++];
	}
      else
	{
	  opts = "";
	}

      if(argc > 1 && argv[oind])
	{
	  pattern = argv[oind++];
	}
      else
	{
	  pattern = "*";
	}

      glob(pattern, globflags, NULL, &gbuf);

      for(i = 0; i < gbuf.gl_pathc; i++)
	{
	  char *path = gbuf.gl_pathv[i];
	  if(stat(path, &st) < 0)
	    {
	      perror(path);
	      continue;
	    }
	  printf("%s", path);
	  if(strchr(opts, 'l'))
	    {
	      printf(" %d", st.st_size);
	    }
	  printf("\n");
	}

      globfree(&gbuf);
    }
  else if(strcmp(argv[0], "cls") == 0)
    {
      printf("\033[2J");
    }
  else if(strcmp(argv[0], "test1") == 0)
    {
      test1();
    }
  else if(strcmp(argv[0], "tabtest") == 0)
    {
      tabtest();
    }
  else if(strcmp(argv[0], "version") == 0)
    {
      showversion();
    }
  else if(strcmp(argv[0], "more") == 0 && is_builtin(argv[0]))
    {
      FILE *f;
      char buf[1024];
      char *p;
      char *fname;
      int nrows;
      int crow = 0;

      if(argc != 2)
	{
	  fprintf(stderr, "Usage: more <file>\n");
	  return -1;
	}

      if((p = getenv("ROWS")) == NULL)
	{
	  fprintf(stderr, "Cannot get ROWS from environment\n");
	  return -1;
	}

      nrows = atoi(p);

      fname = argv[1];
      
      if((f = fopen(fname, "r")) == NULL)
	{
	  perror(fname);
	  return -1;
	}

      while(1)
	{
	  if(fgets(buf, sizeof(buf), f) == NULL)
	    break;

	  printf("%s", buf);

	  crow++;

	  if(crow == nrows - 2)
	    {
	      printf("q to quit, space to continue");
	      fflush(stdout);
	      fgets(buf, sizeof(buf), stdin);
	      if(buf[0] == tolower('q'))
		break;

	      crow = 0;
	    }
	}

      fclose(f);
    }
  else
    {
      int i;
      char *cmdline = NULL;
      int retcode;
      DWORD dwProcId;
      BOOL bWait = TRUE;

      cmdline = malloc(10240);
      cmdline[0] = 0;

      for(i = 0; i < argc; i++)
	{
	  strcat(cmdline, "\"");
	  strcat(cmdline, argv[i]);
	  strcat(cmdline, "\" ");
	}

      if(is_pipeable(argv[0]))
	{
	  char buf[1024];
	  FILE *f = popen(cmdline, "r");
	  while((fgets(buf, sizeof(buf), f)))
	    {
	      fputs(buf, stdout);
	    }
	  pclose(f);
	  retcode = 0; // ??
	}
      else
	{
	  retcode = XCEExecuteProcessA(cmdline, bWait, &dwProcId);
	}

      free(cmdline);

      if(retcode < 0)
	{
	  fprintf(stderr, "OsError: %d\n", GetLastError());
	  return -1;
	}
      else if(retcode != 0)
	{
	  fprintf(stderr, "Retcode: %d\n", retcode);
	  return -1;
	}

      SetForegroundWindow(hMainWnd);
      SetActiveWindow(hMainWnd);
    }

  return 0;
}

void
usage()
{
  XCEShowMessageA("Usage: %s [-f batchfile]", progname);
  exit(1);
}

int
main(int argc, char **argv)
{
  char buf[512];
  char *p;
  int c;

  progname = argv[0];

  while((c = getopt(argc, argv, "f:F:")) != EOF)
    {
      switch(c)
	{
	case 'F':
	case 'f':
	  interactive = 0;
	  batchpath = optarg;
	  if(c == 'F')
	    hidewindow = 1;
	  break;
	default:
	  usage();
	  break;
	}
    }

  readconfig();

  if(batchpath)
    {
      if(freopen(batchpath, "r", stdin) == NULL)
	{
	  XCEShowMessageA("%s: %s", batchpath, strerror(errno));
	  return 1;
	}
    }

  if(hidewindow)
    ShowWindow(hMainWnd, SW_HIDE);

  chdir("/");

  if(interactive)
    {
      clrscr();
      setcursortype(_NORMALCURSOR);
      gotoxy(1,1);
      showversion();

#ifdef USE_READLINE
      termios_init();
      //rl_attempted_completion_function = (CPPFunction *) console_complete;
#endif

    }

  while(1)
    {
      if(interactive)
	{
	  strcpy(buf, readline("# "));
#ifdef USE_READLINE
	  add_history(buf);
#endif
	}
      else
	{
	  if(fgets(buf, sizeof(buf), stdin) == NULL)
	    break;

	  linecount++;
	}

      if((p = strchr(buf, '\n')))
	*p = 0;
      if((p = strchr(buf, '\r')))
	*p = 0;

      if(buf[0] == 0 || buf[0] == '#')
        continue;
      
      if(strcmp(buf, "!") == 0)
	strcpy(buf, lastcommand);
      else
	strcpy(lastcommand, buf);

      substitutevars(buf);
      argv_setargv(buf);

      if(strcmp(a_argv[0], "q") == 0 ||
	 strcmp(a_argv[0], "quit") == 0 ||
	 strcmp(a_argv[0], "exit") == 0)
	{
	  DoExit(0);
	  //break;
	}

      if(handle_command(a_argc, a_argv) < 0 && !interactive)
	{
	  XCEShowMessageA("Error in line %d\n", linecount);
	  break;
	}

      argv_freeargv();
    }

  return 0;
}

static void
argv_append(char *s)
{
  if(a_argc >= a_size - 1)
    {
      // expand argv
      a_size *= 2;
      a_argv = realloc(a_argv, a_size * sizeof(char *));
    }
  // we do not free the glob vector or command-line, so we do not dup
  a_argv[a_argc++] = strdup(s);
  a_argv[a_argc] = NULL;
}

#define GLOBFLAGS (GLOB_ERR | GLOB_NOCHECK | GLOB_NOESCAPE | GLOB_NOMAGIC)

int
argv_setargv(char *command)
{
  int i, res;
  char *p;
  char *s;

  glob_t gbuf;
  memset(&gbuf, 0, sizeof(gbuf));

  a_size = 100;

  a_argc = 0;
  a_argv = malloc(a_size * sizeof(char *));

  s = p = (char *) command;

  for(;;)
    {
      if(*p == 0)
	{
	  if(s != p)
	    {
	      if(globflag)
		{
		  if(glob(s, GLOBFLAGS, NULL, &gbuf) < 0)
		    {
		      fprintf(stderr, "glob error");
		      argv_append(s);
		    }
		  else
		    {
		      for(i = 0; i < gbuf.gl_pathc; i++)
			argv_append(gbuf.gl_pathv[i]);
		    }
		  globfree(&gbuf);
		}
	      else
		{
		  argv_append(s);
		}
	    }
	  break;
	}
      else if(*p == '"')
	{
	  char *start = ++p;

	  while(*p && *p != '"')
	    p++;
	  *p++ = 0;
	  argv_append(start);
	  s = p;
	}
      else if(*p == '\'')
	{
	  char *start = ++p;

	  while(*p && *p != '\'')
	    p++;
	  *p++ = 0;
	  argv_append(start);
	  s = p;
	}
      else if(*p == ' ' || *p == '\t')
	{
	  if(s != p)
	    {
	      *p++ = 0;
	      if(globflag)
		{
		  if(glob(s, GLOBFLAGS, NULL, &gbuf) < 0)
		    {
		      fprintf(stderr, "glob error\n");
		      argv_append(s);
		    }
		  else
		    {
		      for(i = 0; i < gbuf.gl_pathc; i++)
			argv_append(gbuf.gl_pathv[i]);
		    }
		  globfree(&gbuf);
		}
	      else
		{
		  argv_append(s);
		}
	    }
	  while(*p && (*p == '\t' || *p == ' '))
	    p++;
	  s = p;
	}
      else
	{
	  p++;
	}
    }

  return 0;
}

void
argv_freeargv()
{
  char **argv;

  for(argv = a_argv; *argv; argv++)
    free(*argv);

  free(a_argv);
}

int
readconfig()
{
  char buf[512];
  char *p;
  char path[126];
  FILE *f;
  int line = 0;

  if((p = getenv("HOME")) == NULL)
    return -1;
  sprintf(path, "%s/%s", p, ".console");
  if((f = fopen(path, "r")) == NULL)
    {
      sprintf(path, "%s/%s", p, "_console");
      if((f = fopen(path, "r")) == NULL)
	return -1;
    }

  while((fgets(buf, sizeof(buf), f)))
    {
      line++;

      if((p = strrchr(buf, '\n')))
	*p = 0;
      if((p = strrchr(buf, '\r')))
	*p = 0;

      if(buf[0] == '#' || buf[0] == 0)
	continue;

      argv_setargv(buf);

      if(handle_command(a_argc, a_argv) < 0)
	{
	  XCEShowMessageA("Error in init file: line %d", line);
	  exit(1);
	}

      argv_freeargv();
    }

  fclose(f);
  return 0;
}

#ifndef USE_READLINE
char *
readline(char *prompt)
{
  static char buf[1024];
  char *p;

  fputs(prompt, stdout);
  fflush(stdout);
  if(fgets(buf, sizeof(buf), stdin) == NULL)
    return NULL;
  if((p = strrchr(buf, '\n')) != NULL)
    *p = 0;
  return buf;
}
#endif
